/* $Header: /disk/cvs/zaurus/maki/maki.cpp,v 1.3 2003/01/23 13:48:25 robert Exp $ */

/*
 * Copyright (C) 2002 Henry So, Jr.
 *
 * This code may be used and distributed under the GPL.  See LICENSE.
 */

#include "maki.h"

#include <qpe/resource.h>
#include <qpe/config.h>

#include <qvbox.h>
#include <qgrid.h>
#include <qlabel.h>
#include <qaction.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
//#include <qlineedit.h>
#include <qspinbox.h>
#include <qpushbutton.h>
#include <qpe/qpetoolbar.h>
#include <qpe/qpemenubar.h>
#include <qapplication.h>

//#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// min and max constants
static const int min_cell_size = 12;
static const int max_cell_size = 28;
static const int min_tiles = 2;
static const int max_tiles = 7;
static const int def_tiles = 4;
static const int min_cols = 5;
static const int max_cols = 30;
static const int def_cols = 15;
static const int min_rows = 5;
static const int max_rows = 30;
static const int def_rows = 10;

// the color values of the tiles
static const QColor colors[max_tiles + 1] = {
    QColor(0,0,0),          // black (background)
    QColor(255,0,0),        // red
    QColor(255,255,0),      // yellow
    QColor(70,80,255),      // blue
    QColor(0,255,0),        // green
    QColor(255,165,0),      // orange
    QColor(238,130,238),    // violet
    QColor(150,150,150),    // grey
};

// constructor for the main window
MakiMainWindow::MakiMainWindow(QWidget *parent, const char *name)
: QMainWindow(parent, name) {
    setIcon(Resource::loadPixmap("Maki"));

    setToolBarsMovable(FALSE);

    QPEToolBar *toolbar = new QPEToolBar(this);
    toolbar->setHorizontalStretchable(TRUE);
    addToolBar(toolbar);

    QPEMenuBar *menubar = new QPEMenuBar(toolbar);
    menubar->setMargin(0);

    QPopupMenu *game = new QPopupMenu(this);

    QWidget *spacer = new QWidget(toolbar);
    spacer->setBackgroundMode(PaletteButton);
    toolbar->setStretchableWidget(spacer);

    MakiFrame *frame = new MakiFrame(this);
    frame->setFrameShape(QFrame::Box);
    //frame->setFrameShadow(QFrame::Raised);
    frame->setLineWidth(1);

    MakiTable *table = new MakiTable(frame);
    frame->setTable(table);

    setCentralWidget(frame);

    connect(table, SIGNAL(setScore(int)), this, SLOT(setScore(int)));
    connect(table, SIGNAL(setTitle(QString)), this, SLOT(setTitle(QString)));

    QLabel *label = new QLabel(tr("Score: "), toolbar);
    label->setBackgroundMode(PaletteButton);
    _score_label = new QLabel(toolbar);
    _score_label->setBackgroundMode(PaletteButton);
    label = new QLabel(" ", toolbar);
    label->setBackgroundMode(PaletteButton);

    QAction *a = new QAction(tr("Undo / Redo"), QString::null, 0, this, 0);
    connect(a, SIGNAL(activated()), table, SLOT(undo()));
    a->addTo(game);
    a->setEnabled(table->isUndoable());
    connect(table, SIGNAL(setUndoable(bool)), a, SLOT(setEnabled(bool)));

    game->insertSeparator();

    a = new QAction(tr("New"), QString::null, 0, this, 0);
    connect(a, SIGNAL(activated()), table, SLOT(newGame()));
    a->addTo(game);

    a = new QAction(tr("New (Seeded)"), QString::null, 0, this, 0);
    connect(a, SIGNAL(activated()), table, SLOT(newSeededGame()));
    a->addTo(game);

    game->insertSeparator();

    a = new QAction(tr("Preferences"), QString::null, 0, this, 0);
    connect(a, SIGNAL(activated()), table, SLOT(preferences()));
    a->addTo(game);

    menubar->insertItem(tr("Game"), game);

    setScore(table->getScore());
    setTitle(table->getTitle());
}

// slot for setting the score into the toolbar
void
MakiMainWindow::setScore(int score) {
    _score_label->setNum(score);
}

// slot for setting the title into the caption
void
MakiMainWindow::setTitle(QString title) {
    setCaption(title);
}

// handles a resize of the window
void
MakiFrame::resizeEvent(QResizeEvent *e) {
    _table->setAvailableRect(contentsRect());
    QFrame::resizeEvent(e); 
}

// constructor for the game board widget
MakiTable::MakiTable(QWidget *parent, const char *name)
: QScrollView(parent, name), _board(_b), _undo_board(_b + 1), _seed_dialog(0),
        _pref_dialog(0) {
    // setup table view
    setFrameStyle(StyledPanel | Sunken);
    setBackgroundMode(NoBackground);
    setMouseTracking(true);

    readConfig();
}

// destructor for the game board widget
MakiTable::~MakiTable() {
    writeConfig();

    if (_seed_dialog) delete _seed_dialog;
    if (_pref_dialog) delete _pref_dialog;
}

// saves the current state of the application
void
MakiTable::writeConfig() {
    Config cfg("Maki");

    cfg.setGroup("Preferences");
    cfg.writeEntry("Rows", _pref_rows);
    cfg.writeEntry("Columns", _pref_cols);
    cfg.writeEntry("Tiles", _pref_tiles);

    cfg.setGroup("Game");
    cfg.writeEntry("Rows", _board_rows);
    cfg.writeEntry("Columns", _board_cols);
    cfg.writeEntry("Tiles", _board_tiles);
    cfg.writeEntry("Seed", _board_seed);
    cfg.writeEntry("Score", _board_score);
    cfg.writeEntry("Active", _board_active);
    cfg.writeEntry("Undoable", _board_undoable);
    cfg.writeEntry("UndoScore", _undo_score);
    QString board, undo_board;
    const int len = _board_rows * _board_cols;
    for (int i = 0; i < len; i++) {
        board.append(((*_board)[i] & 0x7f) + '0');
        undo_board.append(((*_undo_board)[i] & 0x7f) + '0');
    }
    cfg.writeEntry("Board", board);
    cfg.writeEntry("UndoBoard", undo_board);
}

// loads the previously saved state of the application
void
MakiTable::readConfig() {
    Config cfg("Maki");

    cfg.setGroup("Preferences");
    _pref_rows = cfg.readNumEntry("Rows", def_rows);
    _pref_cols = cfg.readNumEntry("Columns", def_cols);
    _pref_tiles = cfg.readNumEntry("Tiles", def_tiles);

    // sanity check
    if (_pref_tiles > max_tiles) _pref_tiles = max_tiles;

    cfg.setGroup("Game");
    _board_rows = cfg.readNumEntry("Rows", def_rows);
    _board_cols = cfg.readNumEntry("Columns", def_cols);
    _board_tiles = cfg.readNumEntry("Tiles", def_tiles);
    _board_seed = cfg.readNumEntry("Seed", -1);
    _board_score = cfg.readNumEntry("Score", 0);
    _board_active = cfg.readBoolEntry("Active", FALSE);
    _board_undoable = cfg.readBoolEntry("Undoable", FALSE);
    _undo_score = cfg.readNumEntry("UndoScore", -1);
    const QString board = cfg.readEntry("Board", "");
    const QString undo_board = cfg.readEntry("UndoBoard", "");
    const uint len = _board_rows * _board_cols;

    _board->resize(len);
    _undo_board->resize(len);
    bool good = TRUE;
    if (len == board.length()) {
        for (uint i = 0; i < len; i++) {
            char color = char(board.at(i)) - '0';
            if (color < 0 || color > max_tiles) {
                good = FALSE;
                break;
            }
            (*_board)[i] = color;
        }
    }
    else {
        good = FALSE;
    }

    if (!good) {
        _board_active = FALSE;
        _board_score = 0;
        for (uint i = 0; i < len; i++) {
            (*_board)[i] = 0;
        }
    }

    // read the undo board; if misread, board is not undoable
    if (len == undo_board.length()) {
        for (uint i = 0; i < len; i++) {
            char color = char(undo_board.at(i)) - '0';
            if (color < 0 || color > max_tiles) {
                _board_undoable = FALSE;
                break;
            }
            (*_undo_board)[i] = color;
        }
    }
    else {
        _board_undoable = FALSE;
    }

    // if undo score misread, then the board is not undoable
    if (_undo_score < 0) _board_undoable = FALSE;
}

// returns the title string for use in the caption
QString
MakiTable::getTitle() {
    QString title = tr("Q-Maki");
    if (_board_seed < 0) return title;

    // the appends are not translatable
    title += " (";
    title += QString::number(_board_cols);
    title += "x";
    title += QString::number(_board_rows);
    title += "x";
    title += QString::number(_board_tiles);
    title += ": ";
    title += QString::number(_board_seed);
    title += ")";
    return title;
}

// sets the available rectangle for the game board widget
void
MakiTable::setAvailableRect(const QRect &rect) {
    _rect = rect;

    fixCellSize() || fixGeometry();
}

// fixes up the cell size of the tiles and returns whether a change was made
bool
MakiTable::fixCellSize() {
    int new_size = findCellSize();
    if (new_size != _cell_size) {
        viewport()->setUpdatesEnabled(FALSE);

        QFont f = font();
        f.setPixelSize(new_size - 2);
        f.setBold(FALSE);
        setFont(f);

        setCellSize(new_size);

        viewport()->setUpdatesEnabled(TRUE);
        viewport()->repaint(TRUE);

        return TRUE;
    }
    return FALSE;
}

// figures out the cell size depending on the window size and board dimensions
int
MakiTable::findCellSize() {
    const int w = _rect.width() - 1;
    const int h = _rect.height() - 1;

    int cellsize = QMIN(w / _board_cols, h / _board_rows);
    cellsize = QMIN(QMAX(cellsize, min_cell_size), max_cell_size);

    return cellsize;
}

// sets the cell size of the tiles
void
MakiTable::setCellSize(int cellsize) {
    _cell_size = cellsize;
    
    fixGeometry();
}

// fixes up the geometry of the window and returns whether a change was made
bool
MakiTable::fixGeometry() {
    const int w = _rect.width();
    const int h = _rect.height();
    
    int w2 = _cell_size * _board_cols;
    int h2 = _cell_size * _board_rows;

    resizeContents(w2, h2);
    
    const int b = 2;

    bool needHorz = FALSE;
    bool needVert = FALSE;

    if ((w2 + b) > w) {
        needHorz = TRUE;
    }
    if ((h2 + b) > h) {
        needVert = TRUE;
    }
    if (needHorz && !needVert) {
        h2 += horizontalScrollBar()->height();
	if ((h2 + b) > h) {
	    needVert = TRUE;
	}
    } else if (needVert && !needHorz) {
        w2 += verticalScrollBar()->width();
	if ((w2 + b) > h) {
	    needHorz = TRUE;
	}
    }
    setHScrollBarMode(needHorz ? QScrollView::AlwaysOn : QScrollView::AlwaysOff);
    setVScrollBarMode(needVert ? QScrollView::AlwaysOn : QScrollView::AlwaysOff);

    setGeometry(_rect);
/* QRect geom(_rect.x() + QMAX((w - w2) / 2, 0),
            _rect.y() + QMAX((h - h2) / 2, 0),
            QMIN(w, w2 + b), QMIN(h, h2 + b));

    if (geometry() != geom) {
        setGeometry(geom);
        return TRUE;
    }
*/

    return FALSE;
}

// draws the contents of the game board widget
void
MakiTable::drawContents(QPainter * p, int clipx, int clipy, int clipw,
        int cliph ) {
    const int c1 = clipx / _cell_size;
    const int c2 = (clipx + clipw - 1) / _cell_size;
    const int r1 = clipy / _cell_size;
    const int r2 = (clipy + cliph - 1) / _cell_size;
    
    for (int c = c1; c <= c2 ; c++) {
        if (c < 0 || c >= _board_cols) continue;
        for (int r = r1; r <= r2 ; r++) {
            if (r < 0 || r >= _board_rows) continue;

            paintCell(p, r, c);
        }
    }
}

// paints a cell (tile) of the board
void
MakiTable::paintCell(QPainter *p, int r, int c) {
    int x = c * _cell_size;
    int y = r * _cell_size;

    const int s2 = _cell_size - 1;
  
    char tile = tileAt(r, c) & 0x7f;
  
    // draw cell background
    QColor color = colors[tile];
    if (tileAt(r, c) & 0x80) color = color.dark();
    p->setBrush(color);
    p->setPen(NoPen);
    p->drawRect(x, y, _cell_size, _cell_size);
  
    // draw borders
    p->setPen(colorGroup().text());
    if (c < _board_cols - 1) {
        p->drawLine(x + s2, y, x + s2, y + s2); // right border line
    }

    if (r < _board_rows - 1) {
        p->drawLine(x, y + s2, x + s2, y + s2); // bottom boder line
    }
  
    // draw number
    if (tile == 0) return;

    p->setPen(black);
    p->drawText(x, y, s2, s2, AlignHCenter | AlignVCenter,
            QString(QChar(tile - 1 + 'A')));
}

// instructs Qt to update the screen for a given cell (tile)
void
MakiTable::updateCell(int r, int c) {
    updateContents(c * _cell_size, r * _cell_size, _cell_size, _cell_size);
}

// starts a new game
void
MakiTable::startNewGame(int seed) {
    // use rand() to seed random() ... :)
    if (seed < 0) _board_seed = rand();
    else _board_seed = seed;

    srandom(_board_seed);

    // set board variables
    bool maybe_fix = FALSE;
    if (_board_rows != _pref_rows) {
        _board_rows = _pref_rows;
        maybe_fix = TRUE;
    }
    if (_board_cols != _pref_cols) {
        _board_cols = _pref_cols;
        maybe_fix = TRUE;
    }
    if (_board_tiles != _pref_tiles) _board_tiles = _pref_tiles;

    _board_score = 0;
    _board_active = TRUE;
    _board_undoable = FALSE;

    const uint len = _board_rows * _board_cols;
    if (len != _board->size()) _board->resize(len);
    if (len != _undo_board->size()) _undo_board->resize(len);

    // guarantee at least one possible move
    do {
        for (uint i = 0; i < len; i++) {
            (*_board)[i] = (random() % _board_tiles) + 1;
        }
    } while (!moreMovesPossible());

    if (maybe_fix) {
        viewport()->setUpdatesEnabled(FALSE);
        maybe_fix = fixCellSize() || fixGeometry();
        viewport()->setUpdatesEnabled(TRUE);
    }

    viewport()->repaint(TRUE);

    emit setTitle(getTitle());
    emit setScore(_board_score);
    emit setUndoable(_board_undoable);
}

// checks if the game is over
void
MakiTable::checkGameOver() {
    bool cleared;

    if (moreMovesPossible(&cleared)) return;

    if (cleared) {
        // bonus is 2/3 * rows * cols * 10
        _board_score += (_board_rows * _board_cols * 20) / 3;
        emit setScore(_board_score);
    }

    QMessageBox::information(this, tr("Q-Maki"), tr("Game over"));
    _board_active = FALSE;
}

// returns whether more moves are possible on the board
bool
MakiTable::moreMovesPossible(bool *cleared) {
    if (cleared) *cleared = TRUE;

    for (int c = 0; c < _board_cols; c++) {
        for (int r = 0; r < _board_rows; r ++) {
            char tile = tileAt(r, c);
            if (tile == 0) continue;
            if (cleared) *cleared = FALSE;
            if (c + 1 < _board_cols && tile == tileAt(r, c + 1)) return TRUE;
            if (r + 1 < _board_rows && tile == tileAt(r + 1, c)) return TRUE;
        }
    }

    return FALSE;
}

// clears the tag from any tagged tiles
void
MakiTable::clearTags() {
    for (int c = 0; c < _board_cols; c++) {
        for (int r = 0; r < _board_rows; r++) {
            if (tileAt(r, c) & 0x80) {
                tileAt(r, c) &= 0x7f;
                updateCell(r, c);
            }
        }
    }
}

// recursive flood fill algorithm for tagging the tiles around a given tile
int
MakiTable::setTags(int r, int c) {
    const char color = tileAt(r, c);
    if (color == 0) return 0;

    int n = 1;
    tileAt(r, c) |= 0x80;
    updateCell(r, c);
    if (r > 0 && tileAt(r - 1, c) == color) n += setTags(r - 1, c);
    if (c > 0 && tileAt(r, c - 1) == color) n += setTags(r, c - 1);
    if (r + 1 < _board_rows && tileAt(r + 1, c) == color) n+=setTags(r + 1, c);
    if (c + 1 < _board_cols && tileAt(r, c + 1) == color) n+=setTags(r, c + 1);

    return n;
}

// removes the tagged tiles from the board and compresses it
void
MakiTable::removeTaggedTiles() {
    const int maxrow = _board_rows - 1;

    int c = 0;
    // go from left to right, compressing rows and columns simultaneously
    for (int c2 = 0; c2 < _board_cols; c2++) {
        bool empty = FALSE;
        int r2 = maxrow;
        for (int r = maxrow; r >= 0; r--) {
            while (r2 >= 0 && ((tileAt(r2, c2)&0x80) || tileAt(r2, c2) == 0)) {
                r2--;
            }
            if (r2 < 0) {
                if (r == maxrow) {
                    // column empty; compress it
                    empty = TRUE;
                    break;
                }
                undoTileAt(r, c) = 0;
            }
            else {
                undoTileAt(r, c) = tileAt(r2--, c2);
            }
        }
        if (!empty) c++;
    }

    // empty the remaining columns
    for (; c < _board_cols; c++) {
        for (int r = 0; r < _board_rows; r++) {
            undoTileAt(r, c) = 0;
        }
    }

    if (!_board_undoable) {
        _board_undoable = TRUE;
        emit setUndoable(_board_undoable);
    }
    swap_undo_board();

    repaintContents(0, 0, _board_cols * _cell_size, _board_rows * _cell_size,
            FALSE);
}

// swaps the undo board with the current board
void
MakiTable::swap_undo_board() {
    QArray<char> *tmp = _board;
    _board = _undo_board;
    _undo_board = tmp;
}

// undoes/redoes the previous action
void
MakiTable::undo() {
    if (_board_undoable) {
        swap_undo_board();
        clearTags();

        // reactivate the board
        _board_active = true;

        // swap score also
        int tmp = _board_score;
        _board_score = _undo_score;
        _undo_score = tmp;

        emit setScore(_board_score);

        repaintContents(0, 0, _board_cols * _cell_size,
                _board_rows * _cell_size, FALSE);
    }
}

// slot for starting a new game
void
MakiTable::newGame() {
    startNewGame();
}

// slot for starting a new seeded game
void
MakiTable::newSeededGame() {
    if (!_seed_dialog) {
        _seed_dialog = new MakiSeedDialog(this);
    }

    _seed_dialog->setSeed(_board_seed);
    if (_seed_dialog->exec()) {
        startNewGame(_seed_dialog->seed());
    }
}

// slot for bringing up the preferences window
void
MakiTable::preferences() {
    if (!_pref_dialog) {
        _pref_dialog = new MakiPrefDialog(this);
    }

    _pref_dialog->setup(_pref_cols, _pref_rows, _pref_tiles);
    if (_pref_dialog->exec()) {
        _pref_cols = _pref_dialog->cols();
        _pref_rows = _pref_dialog->rows();
        _pref_tiles = _pref_dialog->tiles();
    }
}

// handles a double click
void
MakiTable::contentsMouseDoubleClickEvent(QMouseEvent *e) {
    // handle a double click like two clicks in succession
    contentsMousePressEvent(e);
}

// handles a click
void
MakiTable::contentsMousePressEvent(QMouseEvent *e) {
    if (!_board_active) return;

    int r = e->pos().y() / _cell_size;
    int c = e->pos().x() / _cell_size;

    // sanity check
    if (r < 0 || r >= _board_rows || c < 0 || c >= _board_cols) return;

    if (tileAt(r, c) & 0x80) {
        // clicked on a tagged tile
        removeTaggedTiles();

        _undo_score = _board_score;
        _board_score += (_tagged_count - 2) * (_tagged_count - 2);
        _tagged_count = 0;
        emit setScore(_board_score);

        // check if the game is over
        checkGameOver();
    }
    else {
        // clicked on a non-tagged tile
        clearTags();
        _tagged_count = setTags(r, c);
        if (_tagged_count < 2) {
            tileAt(r, c) &= 0x7f; // untag it
            updateCell(r, c);
            _tagged_count = 0;
        }
    }
}

// constructor for the new seeded game dialog
MakiSeedDialog::MakiSeedDialog(QWidget *parent =0, const char *name =0)
: QDialog(parent, name, TRUE) {
    setCaption(tr("New Game"));
    QGrid *grid = new QGrid(1, this);
    grid->setMinimumWidth(110);
    grid->setMinimumHeight(50);
    grid->setMargin(5);
    grid->setSpacing(5);
    grid->updateGeometry();

    new QLabel(tr("Enter the seed value:"), grid);

    _seed = new QSpinBox(0, 2147483647, 1, grid);
}

// sets the seed value to be displayed in the new seeded game dialog
void
MakiSeedDialog::setSeed(int seed) {
    if (seed < 0) seed = 0;
    else _seed->setValue(seed);
}

// returns the seed entered in the new seeded game dialog
int
MakiSeedDialog::seed() {
    return _seed->text().toInt();
}

// constructor for the preferences dialog
MakiPrefDialog::MakiPrefDialog(QWidget *parent =0,
        const char *name =0)
: QDialog(parent, name, TRUE) {
    setCaption(tr("Preferences"));
    QGrid *grid = new QGrid(2, this);
    grid->setMinimumWidth(110);
    grid->setMinimumHeight(110);
    grid->setMargin(5);
    grid->setSpacing(5);
    grid->updateGeometry();

    new QLabel(tr("Columns:"), grid);
    _cols = new QSpinBox(min_cols, max_cols, 1, grid);
    new QLabel(tr("Rows:"), grid);
    _rows = new QSpinBox(min_rows, max_rows, 1, grid);
    new QLabel(tr("Tiles:"), grid);
    _tiles = new QSpinBox(min_tiles, max_tiles, 1, grid);

    new QWidget(grid);

    QPushButton *button = new QPushButton(tr("Defaults"), grid);
    connect(button, SIGNAL(clicked()), this, SLOT(resetToDefaults()));
}

// sets up the values in the preferences dialog
void
MakiPrefDialog::setup(int cols, int rows, int tiles) {
    _cols->setValue(cols);
    _rows->setValue(rows);
    _tiles->setValue(tiles);
}

// slot for resetting the preferences to the defaults
void
MakiPrefDialog::resetToDefaults() {
    setup(def_cols, def_rows, def_tiles);
}

// returns the column preferences entered in the preferences dialog
int
MakiPrefDialog::cols() {
    return _cols->text().toInt();
}

// returns the row preferences entered in the preferences dialog
int
MakiPrefDialog::rows() {
    return _rows->text().toInt();
}

// returns the tile preferences entered in the preferences dialog
int
MakiPrefDialog::tiles() {
    return _tiles->text().toInt();
}
