/* $Header: /disk/cvs/zaurus/maki/maki.h,v 1.1 2002/07/10 19:21:22 robert Exp $ */

/*
 * Copyright (C) 2002 Henry So, Jr.
 *
 * This code may be used and distributed under the GPL.  See LICENSE.
 */

#ifndef __maki_h
#define __maki_h

#include <qmainwindow.h>
#include <qscrollview.h>
#include <qdialog.h>
#include <qarray.h>

class QLabel;
//class QLineEdit;
class QSpinBox;

// class for the new seeded game dialog
class MakiSeedDialog : public QDialog {
    Q_OBJECT;

public:
    MakiSeedDialog(QWidget *parent =0, const char *name =0);

    int seed();
    void setSeed(int seed);

private:
    QSpinBox *_seed;
};

// class for the preferences dialog
class MakiPrefDialog : public QDialog {
    Q_OBJECT;

public:
    MakiPrefDialog(QWidget *parent =0, const char *name =0);

    void setup(int cols, int rows, int tiles);
    int cols();
    int rows();
    int tiles();

protected slots:
    void resetToDefaults();

private:
    QSpinBox *_cols, *_rows, *_tiles;
};

// class for the game board widget
class MakiTable : public QScrollView {
    Q_OBJECT;

public:
    MakiTable(QWidget *parent =0, const char *name =0);
    ~MakiTable();

    void setAvailableRect(const QRect &rect);
    QString getTitle();
    int getScore() { return _board_score; }
    bool isUndoable() { return _board_undoable; }

protected slots:
    void undo();
    void newGame();
    void newSeededGame();
    void preferences();

signals:
    void setScore(int score);
    void setTitle(QString title);
    void setUndoable(bool undoable);

protected:
    void contentsMouseDoubleClickEvent(QMouseEvent *e);
    void contentsMousePressEvent(QMouseEvent *e);

    void drawContents(QPainter *p, int clipx, int clipy, int clipw, int cliph);

    void readConfig();
    void writeConfig();

    bool fixCellSize();
    int findCellSize();
    void setCellSize(int cellsize);
    bool fixGeometry();

    void updateCell(int y, int x);
    void paintCell(QPainter *p, int r, int c);

    void startNewGame(int seed =-1);
    void checkGameOver();
    bool moreMovesPossible(bool *cleared =0);
    void clearTags();
    int setTags(int r, int c);
    void removeTaggedTiles();

    void swap_undo_board();

    inline char&
    tileAt(int r, int c) { return (*_board)[(c * _board_rows) + r]; }

    inline char&
    undoTileAt(int r, int c) { return (*_undo_board)[(c * _board_rows) + r]; }

private:
    QRect _rect;
    int _cell_size;

    int _pref_rows, _pref_cols, _pref_tiles;
    int _board_rows, _board_cols, _board_tiles, _board_seed, _board_score;
    int _undo_score, _tagged_count;
    QArray<char> _b[2], *_board, *_undo_board;
    MakiSeedDialog *_seed_dialog;
    MakiPrefDialog *_pref_dialog;
    bool _board_active, _board_undoable;
};

// class for the game board frame
class MakiFrame : public QFrame {
    Q_OBJECT;

public:
    MakiFrame(QWidget *parent =0, const char *name =0)
	: QFrame(parent, name) { }
    void setTable(MakiTable *t) { _table = t; }

protected:
    void resizeEvent(QResizeEvent *e);

private:
    MakiTable *_table;
};

// class for the main application window
class MakiMainWindow : public QMainWindow {
    Q_OBJECT;

public:
    MakiMainWindow(QWidget *parent =0, const char *name =0);

public slots:
    void setScore(int score);
    void setTitle(QString title);

private:
    QLabel *_score_label;
};

#endif
